import {
    CalculatorType,
    SectionType,
    Nub,
    Props,
    Observer,
    Session,
    SectionNub,
    acSection,
    ParamDefinition,
    Result
} from "jalhyd";

import { FormulaireElement } from "../elements/formulaire-element";
import { NgParameter, ParamRadioConfig } from "../elements/ngparam";
import { Field } from "../elements/field";
import { StringMap } from "../../stringmap";
import { FormulaireNode } from "../elements/formulaire-node";
import { FieldSet } from "../elements/fieldset";
import { FieldsetContainer } from "../elements/fieldset-container";
import { SelectField } from "../elements/select-field";
import { DeepFieldsetIterator } from "../form-iterator/deep-fieldset-iterator";
import { TopFormulaireElementIterator } from "../form-iterator/top-element-iterator";
import { CalculatorResults } from "../../results/calculator-results";
import { ServiceFactory } from "../../services/service-factory";
import { PabTable } from "../elements/pab-table";
import { SelectEntry } from "../elements/select-entry";

/**
 * classe de base pour tous les formulaires
 */
export abstract class FormulaireDefinition extends FormulaireNode implements Observer {

    /** nom du module de calcul */
    private _calculatorName: string;

    /** Nub courant */
    protected _currentNub: Nub;

    /** propriétés par défaut (lues si _currentNub === undefined ) */
    protected _props = {};

    /** for SectionNubs only */
    protected _defaultNodeType;

    /** aide en ligne pour les résultats */
    protected _resultsHelpLinks: { [key: string]: string };

    /** set to true to definitely gray-out Calculate button  */
    protected _calculateDisabled = false;

    /** fichier de configuration */
    private _jsonConfig: {};

    /** copy of options.resultsHelp read by FormDefinition.parseOptions() */
    public helpLinks: { [key: string]: string };

    constructor() {
        super(undefined);
    }

    // surcharge de FormulaireNode::get:uid()
    // l'UID d'un formulaire est l'UID de son Nub !
    public get uid() {
        if (this._currentNub) {
            return this._currentNub.uid;
        } else {
            throw new Error("Aucun Nub associé au formulaire !");
        }
    }

    public get calculateDisabled(): boolean {
        return this._calculateDisabled;
    }

    public get calculatorType(): CalculatorType {
        const props = this._currentNub === undefined ? this.defaultProperties : (this._currentNub.properties as Props).props;
        return props["calcType"];
    }

    public get nodeType(): SectionType {
        const props = this._currentNub === undefined ? this.defaultProperties : (this._currentNub.properties as Props).props;
        return props["nodeType"];
    }

    public get calculatorName() {
        return this._calculatorName;
    }

    public set calculatorName(name: string) {
        this._calculatorName = name;
        this.notifyObservers({
            "action": "nameChanged",
            "name": name
        }, this);
        // convenient trick to notify param-link components
        ServiceFactory.formulaireService.propagateFormNameChange(this, name);
    }

    public get jsonConfig(): {} {
        return this._jsonConfig;
    }

    public get defaultProperties() {
        return this._props;
    }

    /**
     * Creates a Nub from the given properties, and associates it to the current form
     */
    public initNub(props?: Props) {
        const p = props ? props : new Props(this.defaultProperties);
        this.currentNub = Session.getInstance().createSessionNub(p);

        if (this.currentNub instanceof SectionNub) {
            // add new Section as first child, from given nodeType
            const propsSection = new Props();
            propsSection.setPropValue("calcType", CalculatorType.Section);
            propsSection.setPropValue("nodeType", this._defaultNodeType);
            const section = Session.getInstance().createNub(propsSection);
            this.currentNub.setSection(section as acSection);
        }
    }

    public get currentNub(): Nub {
        return this._currentNub;
    }

    public set currentNub(n: Nub) {
        const nubCalcType = (n.properties as Props).getPropValue("calcType");
        if (this._props["calcType"] !== nubCalcType) {
            throw new Error(
                `Nub ${n.properties["calcType"]} incompatible avec le formulaire ${this._calculatorName} (${this._props["calcType"]})`
            );
        }
        // unsubscribe from old Nub
        if (this._currentNub) {
            this._currentNub.removeObserver(this);
        }
        // replace Nub
        this._currentNub = n;
        // subscribe to new Nub (for result updates)
        this._currentNub.addObserver(this);
    }

    protected deleteNub(sn: Nub) {
        Session.getInstance().deleteNub(sn);
    }

    protected parseOptions(json: {}) {
        const dnt = json["defaultNodeType"];
        if (dnt !== undefined) {
            this._defaultNodeType = SectionType[dnt];
        }
        this._helpLink = json["help"];
        this._resultsHelpLinks = json["resultsHelp"];
        if (json["calculateDisabled"] !== undefined) {
            this._calculateDisabled = json["calculateDisabled"];
        }
    }

    /** called at the end of parseConfig() */
    protected completeParse(firstNotif: boolean = true) {
        this.helpLinks = this._resultsHelpLinks;
    }

    public getOption(json: {}, option: string): any {
        if (json["type"] === "options") {
            return json[option];
        }
    }

    public afterParseFieldset(fs: FieldSet) { }

    public createFieldset(parent: FormulaireNode, json: {}, data?: {}, nub?: Nub): FieldSet {
        const res: FieldSet = new FieldSet(parent);
        if (nub !== undefined) {
            res.setNub(nub, false);
        } else {
            res.setNub(this._currentNub, false);
        }
        parent.kids.push(res);
        return res;
    }

    /**
     * @return true si le SessionNub attaché (ou un de ses enfants) a l'uid donné
     * @param uid id à rechercher
     */
    public hasNubId(uid: string): boolean {
        return (this._currentNub && this._currentNub.uid === uid);
    }

    public moveFieldsetUp(fs: FieldSet) {
    }

    public moveFieldsetDown(fs: FieldSet) {
    }

    public removeFieldset(fs: FieldSet) {
    }

    private parse_fieldset(json: {}, nub?: Nub) {
        const fs = this.createFieldset(this, json, undefined, nub);
        fs.parseConfig(json);
        this.afterParseFieldset(fs);
    }

    private parse_template_container(json: {}, templates: any[]) {
        const fsc: FieldsetContainer = new FieldsetContainer(this);
        fsc.parseConfig(json, templates);
        this.kids.push(fsc);
    }

    private parse_pab_table(json: {}) {
        const tab: PabTable = new PabTable(this);
        tab.parseConfig(json);
        this.kids.push(tab);
    }

    /**
     * 1ère passe d'analyse de la configuration
     */
    public preparseConfig(json: {}) {
        this._jsonConfig = json;

        // analyse des options globales
        // il est utile de le faire avant le reste pour les modules de calcul utilisant
        // des sections (id des selects type de section/variable à calculer)

        for (const conf_index in json) {
            const conf = json[conf_index];
            const type: string = conf["type"];

            if (type === "options") {
                this.parseOptions(conf);
                break;
            }
        }
    }

    /**
     * 2ème passe d'analyse de la configuration
     */
    public parseConfig(json?: {}) {
        if (json !== undefined) {
            this._jsonConfig = json;
        }

        // analyse des éléments du formulaire
        const templates: any[] = [];

        for (const conf_index in this._jsonConfig) {
            const conf = this._jsonConfig[conf_index];
            const type: string = conf["type"];

            switch (type) {
                case "fieldset":
                    this.parse_fieldset(conf);
                    break;

                case "fieldset_template":
                    templates.push(conf);
                    break;

                // options globales
                case "options":
                    break;

                case "template_container":
                    this.parse_template_container(conf, templates);
                    break;

                case "pab_table": // not generic at all
                    this.parse_pab_table(conf);
                    break;

                default:
                    throw new Error(`type d'objet de module de calcul ${type} non pris en charge`);
            }
        }

        this.completeParse();
    }

    public hasParameter(symbol: string): boolean {
        for (const p of this.allFormElements) {
            if (p instanceof NgParameter) {
                if (p.symbol === symbol) {
                    return true;
                }
            }
        }
        return false;
    }

    public getParamFromSymbol(symbol: string): NgParameter {
        for (const p of this.allFormElements) {
            if (p instanceof NgParameter) {
                if (p.symbol === symbol) {
                    return p;
                }
            }
        }
    }

    public getDisplayedParamFromState(st: ParamRadioConfig): NgParameter {
        for (const p of this.allFormElements) {
            if (p.isDisplayed && p instanceof NgParameter) {
                if (p.radioState === st) {
                    return p;
                }
            }
        }
    }

    public getDisplayedParamListFromState(st: ParamRadioConfig): NgParameter[] {
        const res = [];
        for (const p of this.allFormElements) {
            if (p.isDisplayed && p instanceof NgParameter && p.radioState === st) {
                res.push(p);
            }
        }
        return res;
    }

    public getFieldById(id: string): Field {
        const res = this.getFormulaireNodeById(id);
        if (res instanceof Field) {
            return res;
        }
    }

    public getFieldsetById(id: string): FieldSet {
        const res = this.getFormulaireNodeById(id);
        if (res instanceof FieldSet) {
            return res;
        }
    }

    public getParameterValue(symbol: string): number {
        for (const fs of this.allFieldsets) {
            const p = fs.getNodeParameter(symbol);
            if (p !== undefined) {
                switch (p.radioState) {
                    case ParamRadioConfig.FIX:
                        return p.getValue();

                    case ParamRadioConfig.VAR:
                    case ParamRadioConfig.CAL:
                        return undefined;
                }
            }
        }

        throw new Error(`Formulaire.getNodeParameterValue() : pas de paramètre ${symbol} trouvé`);
    }

    protected notifyReset() {
        this.notifyObservers({
            "action": "resetForm"
        }, this);
    }

    /**
     * Forwards Nub's progress updated notification.
     * Used by CalculatorComponent to update progress bar
     */
    protected notifyNubProgressUpdated(sender: any, data: any) {
        this.notifyObservers(data, sender);
    }

    /**
     * réinitialisation du formulaire suite à un changement d'une valeur, d'une option, ... :
     * effacement des résultats, application des dépendances, ...
     */
    public reset() {
        this.resetResults([], undefined, true);
        // prévenir les composants qu'il faut détecter les changements
        this.notifyReset();
    }

    /**
     * retourne la valeur actuellement sélectionnée d'un SelectField
     * @param selectFieldId id du SelectField
     * @returns valeur courante du select sans le préfixe
     */
    public getSelectedValue(selectFieldId: string): any {
        const select: SelectField = <SelectField>this.getFieldById(selectFieldId);
        if (select.multiple) {
            const values: string[] = (select.getValue() as SelectEntry[]).map((se) => se.value);
            return values;
        } else {
            return (select.getValue() as SelectEntry).value;
        }
    }

    /**
     * retourne le label actuellement sélectionnée d'un SelectField
     * @param selectFieldId id du SelectField
     * @returns label courant du select
     */
    public getSelectedLabel(selectFieldId: string): any {
        const select: SelectField = <SelectField>this.getFieldById(selectFieldId);
        if (select.multiple) {
            const labels: string[] = (select.getValue() as SelectEntry[]).map((se) => se.label);
            return labels;
        } else {
            return (select.getValue() as SelectEntry).label;
        }
    }

    public updateLocalisation(lang: string) {
        for (const fe of this.topFormElements) {
            fe.updateLocalisation();
        }
    }

    public isDisplayed(id: string) {
        return (<FormulaireElement>this.getFormulaireNodeById(id)).isDisplayed;
    }

    /**
     * gestion d'un clic sur les radios
     */
    public onRadioClick(info: any) {
        // if mode changed, reset form results
        if (info.oldValueMode !== info.param.valueMode) {
            this.reset();
        }
    }

    /**
     * itère sur tous les FieldSet
     */
    public get allFieldsets(): IterableIterator<FieldSet> {
        return new DeepFieldsetIterator(this);
    }

    /**
     * itère sur tous les FormulaireElement de 1er niveau
    */
    private get topFormElements(): IterableIterator<FormulaireElement> {
        return new TopFormulaireElementIterator(this);
    }

    /**
     * Appelé par CalculatorComponent lorsque le Formulaire est chargé dans la vue,
     * c'est à dire lorsqu'on affiche un module de calcul à l'écran
     */
    public onCalculatorInit() {}

    //  interface Observer

    public update(sender: any, data: any) {
        if (sender instanceof Nub) {
            switch (data.action) {
                case "resultUpdated":
                    // forward Nub results update notification to FormCompute objects
                    this.reaffectResultComponents();
                    break;
            }
        }
    }

    /**
     * Resets the form results, the results panel on screen, the model
     * results, and does the same for all depending modules
     * @param symbol symbol of the parameter whose value change triggered this method
     * @param forceResetAllDependencies if true, even non-calculated non-modified parameters
     *      links will be considered as dependencies @see jalhyd#98
     */
    public resetResults(visited: string[] = [], symbol?: string, forceResetAllDependencies: boolean = false) {
        visited.push(this.currentNub.uid);
        // reset GUI results
        this.resetFormResults();
        // reset model results
        this.currentNub.resetResult();
        // reset the result panels of all forms depending on this one
        ServiceFactory.formulaireService.resetAllDependingFormsResults(this, visited, symbol, forceResetAllDependencies);
    }

    protected abstract compute();

    public abstract get hasResults(): boolean;

    public abstract get results(): CalculatorResults[];

    /**
     * Copies current Nub result into result components for display on page.
     * Should be called every time the Nub result changes.
     * Must be idempotent.
     */
    protected abstract reaffectResultComponents();

    /**
     * retourne un identifiant du paramètre dans le formulaire
     * surchargé dans le cas des ouvrages //
     */
    protected getParameterRefid(p: ParamDefinition) {
        return p.symbol;
    }

    /**
     * Lance le calcul d'un paramètre en déterminant une valeur initiale.
     * Si nécessaire déclenche un calcul en chaîne des modules en amont.
     */
    protected runNubCalc(nub: Nub): Result {
        return nub.CalcSerie();
    }

    protected getComputedParameter(): NgParameter {
        const cpd = this.currentNub.calculatedParam;
        let ngparam: NgParameter;
        if (cpd !== undefined) {
            ngparam = this.getParamFromSymbol(cpd.symbol);
            if (ngparam === undefined) { // calculated parameter is not displayed on screen
                ngparam = new NgParameter(cpd, this);
            }
        }
        return ngparam;
    }

    public getVariatedParameters(): NgParameter[] {
        let res: NgParameter[] = [];
        // find variated local parameters
        res = this.getDisplayedParamListFromState(ParamRadioConfig.VAR);
        // add variated linked parameters
        const pms = this.getDisplayedParamListFromState(ParamRadioConfig.LINK);
        for (const p of pms) {
            if (p.paramDefinition.hasMultipleValues) {
                res.push(p);
            }
        }
        return res;
    }

    /**
     * Triggers computation of the Nub, updates form results
     */
    public doCompute() {
        // calculate module
        this.compute();
        // refresh results
        this.notifyObservers({
            "action": "resultsUpdated",
        }, this);
    }

    public resetFormResults() {}
}
