import { PlottableData } from "./plottable-data";
import { PabResults } from "./pab-results";
import { ChartType } from "./chart-type";
import { ServiceFactory } from "../services/service-factory";

export class PlottablePabResults implements PlottableData {

    public graphType: ChartType = ChartType.Scatter;
    public chartX: string;
    public chartY: string;

    protected pabResults: PabResults;

    public constructor(pabResults?: PabResults) {
        if (pabResults) {
            this.setPabResults(pabResults);
        }
        // axes par défaut
        this.chartX = this.chartX || "CLOISON";
        this.chartY = this.chartY || "YMOY";
    }

    /** reaffect pabResults, for ex. when objet was contructed with empty pabResults */
    public setPabResults(pabResults: PabResults) {
        this.pabResults = pabResults;
    }

    /**
     * Returns the label to display, for an element of getAvailableChartAxis()
     * @param symbol parameter / result symbol (ex: "Q")
     */
    public getChartAxisLabel(symbol: string): string {
        if (symbol === "x") { // specific case for wall abscissa
            return ServiceFactory.i18nService.localizeText("INFO_LIB_ABSCISSE_CLOISON");
        } else {
            return this.pabResults.headers[this.pabResults.columns.indexOf(symbol)];
        }
    }

    public expandLabelFromSymbol(symbol: string): string {
        return symbol;
    }

    /**
     * Returns a list of plottable parameters / result elements, that can be defined
     * as X or Y chart axis
     */
    public getAvailableChartAxis(): string[] {
        const axis = [];
        axis.push("x"); // wall abscissa
        for (const c of this.pabResults.columns) {
            if (c.indexOf("ENUM_") === -1) { // ENUM variables are not plottable
                axis.push(c);
            }
        }
        return axis;
    }

    public getAvailableXAxis(): string[] {
        return this.getAvailableChartAxis();
    }

    public getAvailableYAxis(): string[] {
        return this.getAvailableChartAxis();
    }

    // just to implement interface
    public getVariatingParametersSymbols(): string[] {
        return [];
    }

    /**
     * Returns the series of values for the required symbol
     * @param symbol parameter / result symbol (ex: "Q")
     */
    public getValuesSeries(symbol: string): number[] {
        const data: number[] = [];
        const pr = this.pabResults;
        const l = this.pabResults.cloisonsResults.length;
        // when a parameter is variating, index of the variating parameter
        // values to build the data from
        const vi = this.pabResults.variableIndex;

        if (this.pabResults.iterationHasError(vi)) {
            return [];
        }

        switch (symbol) {
            case "CLOISON":
                data.push(undefined);
                for (let i = 0; i <= l; i++) { // <= for one extra step (downwall)
                    data.push(i + 1);
                }
                break;

            case "DH":
            case "ZRAM":
            case "Q":
                data.push(undefined);
                for (let i = 0; i < l; i++) {
                    const er = pr.cloisonsResults[i].resultElements[vi].getValue(symbol);
                    data.push(er);
                }
                const zrAval = pr.cloisonAvalResults.resultElements[vi].getValue(symbol);
                data.push(zrAval);
                break;

            case "Z":
                for (let i = 0; i < l; i++) {
                    data.push(pr.cloisonsResults[i].resultElements[vi].vCalc);
                }
                data.push(pr.cloisonAvalResults.resultElements[vi].vCalc);
                data.push(pr.Z2[vi]);
                break;

            case "PV":
            case "YMOY":
            case "ZRMB":
            case "QA":
                data.push(undefined);
                for (let i = 0; i < l; i++) {
                    const er = pr.cloisonsResults[i].resultElements[vi].getValue(symbol);
                    data.push(er);
                }
                data.push(undefined);
                break;

            case "x": // wall abscissa
                data.push(undefined);
                for (let i = 0; i < l; i++) {
                    const er = pr.cloisonsResults[i].resultElements[vi].getValue(symbol);
                    data.push(er);
                }
                const erXdw = pr.cloisonAvalResults.resultElements[vi].getValue(symbol);
                data.push(erXdw);
                break;
        }

        return data;
    }
}
